# -*- coding: utf-8 -*-
# _blobs.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Clientside BlobBackend Storage.
"""

from urlparse import urljoin

import binascii
import os
import base64

from io import BytesIO
from functools import partial

from twisted.logger import Logger
from twisted.enterprise import adbapi
from twisted.internet import defer
from twisted.web.client import FileBodyProducer

import treq

from leap.soledad.common.errors import SoledadError
from leap.common.files import mkdir_p

from .._document import BlobDoc
from .._crypto import DocInfo
from .._crypto import BlobEncryptor
from .._crypto import BlobDecryptor
from .._http import HTTPClient
from .._pipes import TruncatedTailPipe
from .._pipes import PreamblePipe

from . import pragmas
from . import sqlcipher


logger = Logger()
FIXED_REV = 'ImmutableRevision'  # Blob content is immutable


class BlobAlreadyExistsError(SoledadError):
    pass


class ConnectionPool(adbapi.ConnectionPool):

    def insertAndGetLastRowid(self, *args, **kwargs):
        """
        Execute an SQL query and return the last rowid.

        See: https://sqlite.org/c3ref/last_insert_rowid.html
        """
        return self.runInteraction(
            self._insertAndGetLastRowid, *args, **kwargs)

    def _insertAndGetLastRowid(self, trans, *args, **kw):
        trans.execute(*args, **kw)
        return trans.lastrowid

    def blob(self, table, column, irow, flags):
        """
        Open a BLOB for incremental I/O.

        Return a handle to the BLOB that would be selected by:

          SELECT column FROM table WHERE rowid = irow;

        See: https://sqlite.org/c3ref/blob_open.html

        :param table: The table in which to lookup the blob.
        :type table: str
        :param column: The column where the BLOB is located.
        :type column: str
        :param rowid: The rowid of the BLOB.
        :type rowid: int
        :param flags: If zero, BLOB is opened for read-only. If non-zero,
                      BLOB is opened for RW.
        :type flags: int

        :return: A BLOB handle.
        :rtype: pysqlcipher.dbapi.Blob
        """
        return self.runInteraction(self._blob, table, column, irow, flags)

    def _blob(self, trans, table, column, irow, flags):
        # TODO: should not use transaction private variable here
        handle = trans._connection.blob(table, column, irow, flags)
        return handle


def check_http_status(code):
    if code == 409:
        raise BlobAlreadyExistsError()
    elif code != 200:
        raise SoledadError("Server Error")


class DecrypterBuffer(object):

    def __init__(self, blob_id, secret, tag):
        self.doc_info = DocInfo(blob_id, FIXED_REV)
        self.secret = secret
        self.tag = tag
        self.preamble_pipe = PreamblePipe(self._make_decryptor)

    def _make_decryptor(self, preamble):
        self.decrypter = BlobDecryptor(
            self.doc_info, preamble,
            secret=self.secret,
            armor=False,
            start_stream=False,
            tag=self.tag)
        return TruncatedTailPipe(self.decrypter, tail_size=len(self.tag))

    def write(self, data):
        self.preamble_pipe.write(data)

    def close(self):
        real_size = self.decrypter.decrypted_content_size
        return self.decrypter._end_stream(), real_size


class BlobManager(object):
    """
    Ideally, the decrypting flow goes like this:

    - GET a blob from remote server.
    - Decrypt the preamble
    - Allocate a zeroblob in the sqlcipher sink
    - Mark the blob as unusable (ie, not verified)
    - Decrypt the payload incrementally, and write chunks to sqlcipher
      ** Is it possible to use a small buffer for the aes writer w/o
      ** allocating all the memory in openssl?
    - Finalize the AES decryption
    - If preamble + payload verifies correctly, mark the blob as usable

    """

    def __init__(
            self, local_path, remote, key, secret, user, token=None,
            cert_file=None):
        if local_path:
            mkdir_p(os.path.dirname(local_path))
            self.local = SQLiteBlobBackend(local_path, key)
        self.remote = remote
        self.secret = secret
        self.user = user
        self._client = HTTPClient(user, token, cert_file)

    def close(self):
        if hasattr(self, 'local') and self.local:
            return self.local.close()

    @defer.inlineCallbacks
    def remote_list(self):
        uri = urljoin(self.remote, self.user + '/')
        data = yield self._client.get(uri)
        defer.returnValue((yield data.json()))

    def local_list(self):
        return self.local.list()

    @defer.inlineCallbacks
    def send_missing(self):
        our_blobs = yield self.local_list()
        server_blobs = yield self.remote_list()
        missing = [b_id for b_id in our_blobs if b_id not in server_blobs]
        logger.info("Amount of documents missing on server: %s" % len(missing))
        # TODO: Send concurrently when we are able to stream directly from db
        for blob_id in missing:
            fd = yield self.local.get(blob_id)
            logger.info("Upload local blob: %s" % blob_id)
            yield self._encrypt_and_upload(blob_id, fd)

    @defer.inlineCallbacks
    def fetch_missing(self):
        # TODO: Use something to prioritize user requests over general new docs
        our_blobs = yield self.local_list()
        server_blobs = yield self.remote_list()
        docs_we_want = [b_id for b_id in server_blobs if b_id not in our_blobs]
        logger.info("Fetching new docs from server: %s" % len(docs_we_want))
        # TODO: Fetch concurrently when we are able to stream directly into db
        for blob_id in docs_we_want:
            logger.info("Fetching new doc: %s" % blob_id)
            yield self.get(blob_id)

    @defer.inlineCallbacks
    def put(self, doc, size):
        if (yield self.local.exists(doc.blob_id)):
            error_message = "Blob already exists: %s" % doc.blob_id
            raise BlobAlreadyExistsError(error_message)
        fd = doc.blob_fd
        # TODO this is a tee really, but ok... could do db and upload
        # concurrently. not sure if we'd gain something.
        yield self.local.put(doc.blob_id, fd, size=size)
        # In fact, some kind of pipe is needed here, where each write on db
        # handle gets forwarded into a write on the connection handle
        fd = yield self.local.get(doc.blob_id)
        yield self._encrypt_and_upload(doc.blob_id, fd)

    @defer.inlineCallbacks
    def get(self, blob_id):
        local_blob = yield self.local.get(blob_id)
        if local_blob:
            logger.info("Found blob in local database: %s" % blob_id)
            defer.returnValue(local_blob)

        result = yield self._download_and_decrypt(blob_id)

        if not result:
            defer.returnValue(None)
        blob, size = result

        if blob:
            logger.info("Got decrypted blob of type: %s" % type(blob))
            blob.seek(0)
            yield self.local.put(blob_id, blob, size=size)
            defer.returnValue((yield self.local.get(blob_id)))
        else:
            # XXX we shouldn't get here, but we will...
            # lots of ugly error handling possible:
            # 1. retry, might be network error
            # 2. try later, maybe didn't finished streaming
            # 3.. resignation, might be error while verifying
            logger.error('sorry, dunno what happened')

    @defer.inlineCallbacks
    def _encrypt_and_upload(self, blob_id, fd):
        # TODO ------------------------------------------
        # this is wrong, is doing 2 stages.
        # the crypto producer can be passed to
        # the uploader and react as data is written.
        # try to rewrite as a tube: pass the fd to aes and let aes writer
        # produce data to the treq request fd.
        # ------------------------------------------------
        logger.info("Staring upload of blob: %s" % blob_id)
        doc_info = DocInfo(blob_id, FIXED_REV)
        uri = urljoin(self.remote, self.user + "/" + blob_id)
        crypter = BlobEncryptor(doc_info, fd, secret=self.secret,
                                armor=False)
        fd = yield crypter.encrypt()
        response = yield self._client.put(uri, data=fd)
        check_http_status(response.code)
        logger.info("Finished upload: %s" % (blob_id,))

    @defer.inlineCallbacks
    def _download_and_decrypt(self, blob_id):
        logger.info("Staring download of blob: %s" % blob_id)
        # TODO this needs to be connected in a tube
        uri = urljoin(self.remote, self.user + '/' + blob_id)
        data = yield self._client.get(uri)

        if data.code == 404:
            logger.warn("Blob not found in server: %s" % blob_id)
            defer.returnValue(None)
        elif not data.headers.hasHeader('Tag'):
            logger.error("Server didn't send a tag header for: %s" % blob_id)
            defer.returnValue(None)
        tag = data.headers.getRawHeaders('Tag')[0]
        tag = base64.urlsafe_b64decode(tag)
        buf = DecrypterBuffer(blob_id, self.secret, tag)

        # incrementally collect the body of the response
        yield treq.collect(data, buf.write)
        fd, size = buf.close()
        logger.info("Finished download: (%s, %d)" % (blob_id, size))
        defer.returnValue((fd, size))

    @defer.inlineCallbacks
    def delete(self, blob_id):
        logger.info("Staring deletion of blob: %s" % blob_id)
        yield self._delete_from_remote(blob_id)
        if (yield self.local.exists(blob_id)):
            yield self.local.delete(blob_id)

    def _delete_from_remote(self, blob_id):
        # TODO this needs to be connected in a tube
        uri = urljoin(self.remote, self.user + '/' + blob_id)
        return self._client.delete(uri)


class SQLiteBlobBackend(object):

    def __init__(self, path, key=None):
        self.path = os.path.abspath(
            os.path.join(path, 'soledad_blob.db'))
        mkdir_p(os.path.dirname(self.path))
        if not key:
            raise ValueError('key cannot be None')
        backend = 'pysqlcipher.dbapi2'
        opts = sqlcipher.SQLCipherOptions(
            '/tmp/ignored', binascii.b2a_hex(key),
            is_raw_key=True, create=True)
        pragmafun = partial(pragmas.set_init_pragmas, opts=opts)
        openfun = _sqlcipherInitFactory(pragmafun)

        self.dbpool = ConnectionPool(
            backend, self.path, check_same_thread=False, timeout=5,
            cp_openfun=openfun, cp_min=1, cp_max=2, cp_name='blob_pool')

    def close(self):
        from twisted._threads import AlreadyQuit
        try:
            self.dbpool.close()
        except AlreadyQuit:
            pass

    @defer.inlineCallbacks
    def put(self, blob_id, blob_fd, size=None):
        logger.info("Saving blob in local database...")
        insert = 'INSERT INTO blobs (blob_id, payload) VALUES (?, zeroblob(?))'
        irow = yield self.dbpool.insertAndGetLastRowid(insert, (blob_id, size))
        handle = yield self.dbpool.blob('blobs', 'payload', irow, 1)
        blob_fd.seek(0)
        # XXX I have to copy the buffer here so that I'm able to
        # return a non-closed file to the caller (blobmanager.get)
        # FIXME should remove this duplication!
        # have a look at how treq does cope with closing the handle
        # for uploading a file
        producer = FileBodyProducer(blob_fd)
        done = yield producer.startProducing(handle)
        logger.info("Finished saving blob in local database.")
        defer.returnValue(done)

    @defer.inlineCallbacks
    def get(self, blob_id):
        # TODO we can also stream the blob value using sqlite
        # incremental interface for blobs - and just return the raw fd instead
        select = 'SELECT payload FROM blobs WHERE blob_id = ?'
        result = yield self.dbpool.runQuery(select, (blob_id,))
        if result:
            defer.returnValue(BytesIO(str(result[0][0])))

    @defer.inlineCallbacks
    def list(self):
        query = 'select blob_id from blobs'
        result = yield self.dbpool.runQuery(query)
        if result:
            defer.returnValue([b_id[0] for b_id in result])
        else:
            defer.returnValue([])

    @defer.inlineCallbacks
    def exists(self, blob_id):
        query = 'SELECT blob_id from blobs WHERE blob_id = ?'
        result = yield self.dbpool.runQuery(query, (blob_id,))
        defer.returnValue(bool(len(result)))

    def delete(self, blob_id):
        query = 'DELETE FROM blobs WHERE blob_id = ?'
        return self.dbpool.runQuery(query, (blob_id,))


def _init_blob_table(conn):
    maybe_create = (
        "CREATE TABLE IF NOT EXISTS "
        "blobs ("
        "blob_id PRIMARY KEY, "
        "payload BLOB)")
    conn.execute(maybe_create)


def _sqlcipherInitFactory(fun):
    def _initialize(conn):
        fun(conn)
        _init_blob_table(conn)
    return _initialize


#
# testing facilities
#

@defer.inlineCallbacks
def testit(reactor):
    # configure logging to stdout
    from twisted.python import log
    import sys
    log.startLogging(sys.stdout)

    # parse command line arguments
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--url', default='http://localhost:9000/')
    parser.add_argument('--path', default='/tmp/blobs')
    parser.add_argument('--secret', default='secret')
    parser.add_argument('--uuid', default='user')
    parser.add_argument('--token', default=None)
    parser.add_argument('--cert-file', default='')

    subparsers = parser.add_subparsers(help='sub-command help', dest='action')

    # parse upload command
    parser_upload = subparsers.add_parser(
        'upload', help='upload blob and bypass local db')
    parser_upload.add_argument('payload')
    parser_upload.add_argument('blob_id')

    # parse download command
    parser_download = subparsers.add_parser(
        'download', help='download blob and bypass local db')
    parser_download.add_argument('blob_id')
    parser_download.add_argument('--output-file', default='/tmp/incoming-file')

    # parse put command
    parser_put = subparsers.add_parser(
        'put', help='put blob in local db and upload')
    parser_put.add_argument('payload')
    parser_put.add_argument('blob_id')

    # parse get command
    parser_get = subparsers.add_parser(
        'get', help='get blob from local db, get if needed')
    parser_get.add_argument('blob_id')

    # parse delete command
    parser_get = subparsers.add_parser(
        'delete', help='delete blob from local and remote db')
    parser_get.add_argument('blob_id')

    # parse list command
    parser_get = subparsers.add_parser(
        'list', help='list local and remote blob ids')

    # parse send_missing command
    parser_get = subparsers.add_parser(
        'send_missing', help='send all pending upload blobs')

    # parse send_missing command
    parser_get = subparsers.add_parser(
        'fetch_missing', help='fetch all new server blobs')

    # parse arguments
    args = parser.parse_args()

    # TODO convert these into proper unittests

    def _manager():
        mkdir_p(os.path.dirname(args.path))
        manager = BlobManager(
            args.path, args.url,
            'A' * 32, args.secret,
            args.uuid, args.token, args.cert_file)
        return manager

    @defer.inlineCallbacks
    def _upload(blob_id, payload):
        logger.info(":: Starting upload only: %s" % str((blob_id, payload)))
        manager = _manager()
        with open(payload, 'r') as fd:
            yield manager._encrypt_and_upload(blob_id, fd)
        logger.info(":: Finished upload only: %s" % str((blob_id, payload)))

    @defer.inlineCallbacks
    def _download(blob_id):
        logger.info(":: Starting download only: %s" % blob_id)
        manager = _manager()
        result = yield manager._download_and_decrypt(blob_id)
        logger.info(":: Result of download: %s" % str(result))
        if result:
            fd, _ = result
            with open(args.output_file, 'w') as f:
                logger.info(":: Writing data to %s" % args.output_file)
                f.write(fd.read())
        logger.info(":: Finished download only: %s" % blob_id)

    @defer.inlineCallbacks
    def _put(blob_id, payload):
        logger.info(":: Starting full put: %s" % blob_id)
        manager = _manager()
        size = os.path.getsize(payload)
        with open(payload) as fd:
            doc = BlobDoc(fd, blob_id)
            result = yield manager.put(doc, size=size)
        logger.info(":: Result of put: %s" % str(result))
        logger.info(":: Finished full put: %s" % blob_id)

    @defer.inlineCallbacks
    def _get(blob_id):
        logger.info(":: Starting full get: %s" % blob_id)
        manager = _manager()
        fd = yield manager.get(blob_id)
        if fd:
            logger.info(":: Result of get: " + fd.getvalue())
        logger.info(":: Finished full get: %s" % blob_id)

    @defer.inlineCallbacks
    def _delete(blob_id):
        logger.info(":: Starting deletion of: %s" % blob_id)
        manager = _manager()
        yield manager.delete(blob_id)
        logger.info(":: Finished deletion of: %s" % blob_id)

    @defer.inlineCallbacks
    def _list():
        logger.info(":: Listing local blobs")
        manager = _manager()
        local_list = yield manager.local_list()
        logger.info(":: Local list: %s" % local_list)
        logger.info(":: Listing remote blobs")
        remote_list = yield manager.remote_list()
        logger.info(":: Remote list: %s" % remote_list)

    @defer.inlineCallbacks
    def _send_missing():
        logger.info(":: Sending local pending upload docs")
        manager = _manager()
        yield manager.send_missing()
        logger.info(":: Finished sending missing docs")

    @defer.inlineCallbacks
    def _fetch_missing():
        logger.info(":: Fetching remote new docs")
        manager = _manager()
        yield manager.fetch_missing()
        logger.info(":: Finished fetching new docs")

    if args.action == 'upload':
        yield _upload(args.blob_id, args.payload)
    elif args.action == 'download':
        yield _download(args.blob_id)
    elif args.action == 'put':
        yield _put(args.blob_id, args.payload)
    elif args.action == 'get':
        yield _get(args.blob_id)
    elif args.action == 'delete':
        yield _delete(args.blob_id)
    elif args.action == 'list':
        yield _list()
    elif args.action == 'send_missing':
        yield _send_missing()
    elif args.action == 'fetch_missing':
        yield _fetch_missing()


if __name__ == '__main__':
    from twisted.internet.task import react
    react(testit)



# -*- coding: utf-8 -*-
# send.py
# Copyright (C) 2015 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import json

from twisted.internet import defer

from leap.soledad.common.log import getLogger
from leap.soledad.client.events import emit_async
from leap.soledad.client.events import SOLEDAD_SYNC_SEND_STATUS
from leap.soledad.client.http_target.support import RequestBody
from .send_protocol import DocStreamProducer

logger = getLogger(__name__)


class HTTPDocSender(object):
    """
    Handles Document uploading from Soledad server, using HTTP as transport.
    They need to be encrypted and metadata prepared before sending.
    """

    # The uuid of the local replica.
    # Any class inheriting from this one should provide a meaningful attribute
    # if the sync status event is meant to be used somewhere else.

    uuid = 'undefined'
    userid = 'undefined'

    @defer.inlineCallbacks
    def _send_docs(self, docs_by_generation, last_known_generation,
                   last_known_trans_id, sync_id):

        if not docs_by_generation:
            defer.returnValue([None, None])

        # add remote replica metadata to the request
        body = RequestBody(
            last_known_generation=last_known_generation,
            last_known_trans_id=last_known_trans_id,
            sync_id=sync_id,
            ensure=self._ensure_callback is not None)
        result = yield self._send_batch(body, docs_by_generation)
        response_dict = json.loads(result)[0]
        gen_after_send = response_dict['new_generation']
        trans_id_after_send = response_dict['new_transaction_id']
        defer.returnValue([gen_after_send, trans_id_after_send])

    @defer.inlineCallbacks
    def _send_batch(self, body, docs):
        total, calls = len(docs), []
        for i, entry in enumerate(docs):
            calls.append((self._prepare_one_doc,
                         entry, body, i + 1, total))
        result = yield self._send_request(body, calls)
        _emit_send_status(self.uuid, body.consumed, total)

        defer.returnValue(result)

    def _send_request(self, body, calls):
        return self._http_request(
            self._url,
            method='POST',
            body=(body, calls),
            content_type='application/x-soledad-sync-put',
            body_producer=DocStreamProducer)

    @defer.inlineCallbacks
    def _prepare_one_doc(self, entry, body, idx, total):
        get_doc_call, gen, trans_id = entry
        doc, content = yield self._encrypt_doc(get_doc_call)
        body.insert_info(
            id=doc.doc_id, rev=doc.rev, content=content, gen=gen,
            trans_id=trans_id, number_of_docs=total,
            doc_idx=idx)
        _emit_send_status(self.uuid, body.consumed, total)

    @defer.inlineCallbacks
    def _encrypt_doc(self, get_doc_call):
        f, args, kwargs = get_doc_call
        doc = yield f(*args, **kwargs)
        if doc.is_tombstone():
            defer.returnValue((doc, None))
        else:
            content = yield self._crypto.encrypt_doc(doc)
            defer.returnValue((doc, content))


def _emit_send_status(user_data, idx, total):
    content = {'sent': idx, 'total': total}
    emit_async(SOLEDAD_SYNC_SEND_STATUS, user_data, content)

    msg = "%d/%d" % (idx, total)
    logger.debug("Sync send status: %s" % msg)



# -*- coding: utf-8 -*-
# _secrets/storage.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import six.moves.urllib.parse as urlparse

from hashlib import sha256

from leap.soledad.common import SHARED_DB_NAME
from leap.soledad.common.log import getLogger

from leap.soledad.client.shared_db import SoledadSharedDatabase
from leap.soledad.client._document import Document
from leap.soledad.client._secrets.util import emit, UserDataMixin


logger = getLogger(__name__)


class SecretsStorage(UserDataMixin):

    def __init__(self, soledad):
        self._soledad = soledad
        self._shared_db = self._soledad.shared_db or self._init_shared_db()
        self.__remote_doc = None

    @property
    def _creds(self):
        uuid = self._soledad.uuid
        token = self._soledad.token
        return {'token': {'uuid': uuid, 'token': token}}

    #
    # local storage
    #

    def load_local(self):
        path = self._soledad.secrets_path
        logger.info("trying to load secrets from disk: %s" % path)
        try:
            with open(path, 'r') as f:
                encrypted = json.loads(f.read())
            logger.info("secrets loaded successfully from disk")
            return encrypted
        except IOError:
            logger.warn("secrets not found in disk")
        return None

    def save_local(self, encrypted):
        path = self._soledad.secrets_path
        json_data = json.dumps(encrypted)
        with open(path, 'w') as f:
            f.write(json_data)

    #
    # remote storage
    #

    def _init_shared_db(self):
        url = urlparse.urljoin(self._soledad.server_url, SHARED_DB_NAME)
        creds = self._creds
        db = SoledadSharedDatabase.open_database(url, creds)
        return db

    def _remote_doc_id(self):
        passphrase = self._soledad.passphrase.encode('utf8')
        uuid = self._soledad.uuid
        text = '%s%s' % (passphrase, uuid)
        digest = sha256(text).hexdigest()
        return digest

    @property
    def _remote_doc(self):
        if not self.__remote_doc and self._shared_db:
            doc = self._get_remote_doc()
            self.__remote_doc = doc
        return self.__remote_doc

    @emit('downloading')
    def _get_remote_doc(self):
        logger.info('trying to load secrets from server...')
        doc = self._shared_db.get_doc(self._remote_doc_id())
        if doc:
            logger.info('secrets loaded successfully from server')
        else:
            logger.warn('secrets not found in server')
        return doc

    def load_remote(self):
        doc = self._remote_doc
        if not doc:
            return None
        encrypted = doc.content
        return encrypted

    @emit('uploading')
    def save_remote(self, encrypted):
        doc = self._remote_doc
        if not doc:
            doc = Document(doc_id=self._remote_doc_id())
        doc.content = encrypted
        db = self._shared_db
        if not db:
            logger.warn('no shared db found')
            return
        db.put_doc(doc)



# -*- coding: utf-8 -*-
# interfaces.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Interfaces used by the Soledad Client.
"""
from zope.interface import Interface, Attribute

#
# Plugins
#


class ISoledadPostSyncPlugin(Interface):
    """
    I implement the minimal methods and attributes for a plugin that can be
    called after a soledad synchronization has ended.
    """

    def process_received_docs(self, doc_id_list):
        """
        Do something with the passed list of doc_ids received after the last
        sync.

        :param doc_id_list: a list of strings for the received doc_ids
        """

    watched_doc_types = Attribute("""
        a tuple of the watched doc types for this plugin. So far, the
        `doc-types` convention is just the preffix of the doc_id, which is
        basically its first character, followed by a dash. So, for instance,
        `M-` is used for meta-docs in mail, and `F-` is used for flag-docs in
        mail. For now there's no central register of all the doc-types
        used.""")


#
# Soledad storage
#

class ILocalStorage(Interface):
    """
    I implement core methods for the u1db local storage of documents and
    indexes.
    """
    local_db_path = Attribute(
        "The path for the local database replica")
    local_db_file_name = Attribute(
        "The name of the local SQLCipher U1DB database file")
    uuid = Attribute("The user uuid")
    default_prefix = Attribute(
        "Prefix for default values for path")

    def put_doc(self, doc):
        """
        Update a document in the local encrypted database.

        :param doc: the document to update
        :type doc: Document

        :return:
            a deferred that will fire with the new revision identifier for
            the document
        :rtype: Deferred
        """

    def delete_doc(self, doc):
        """
        Delete a document from the local encrypted database.

        :param doc: the document to delete
        :type doc: Document

        :return:
            a deferred that will fire with ...
        :rtype: Deferred
        """

    def get_doc(self, doc_id, include_deleted=False):
        """
        Retrieve a document from the local encrypted database.

        :param doc_id: the unique document identifier
        :type doc_id: str
        :param include_deleted:
            if True, deleted documents will be returned with empty content;
            otherwise asking for a deleted document will return None
        :type include_deleted: bool

        :return:
            A deferred that will fire with the document object, containing a
            Document, or None if it could not be found
        :rtype: Deferred
        """

    def get_docs(self, doc_ids, check_for_conflicts=True,
                 include_deleted=False):
        """
        Get the content for many documents.

        :param doc_ids: a list of document identifiers
        :type doc_ids: list
        :param check_for_conflicts: if set False, then the conflict check will
            be skipped, and 'None' will be returned instead of True/False
        :type check_for_conflicts: bool

        :return:
            A deferred that will fire with an iterable giving the Document
            object for each document id in matching doc_ids order.
        :rtype: Deferred
        """

    def get_all_docs(self, include_deleted=False):
        """
        Get the JSON content for all documents in the database.

        :param include_deleted: If set to True, deleted documents will be
                                returned with empty content. Otherwise deleted
                                documents will not be included in the results.
        :return:
            A deferred that will fire with (generation, [Document]): that is,
            the current generation of the database, followed by a list of all
            the documents in the database.
        :rtype: Deferred
        """

    def create_doc(self, content, doc_id=None):
        """
        Create a new document in the local encrypted database.

        :param content: the contents of the new document
        :type content: dict
        :param doc_id: an optional identifier specifying the document id
        :type doc_id: str

        :return:
            A deferred tht will fire with the new document (Document
            instance).
        :rtype: Deferred
        """

    def create_doc_from_json(self, json, doc_id=None):
        """
        Create a new document.

        You can optionally specify the document identifier, but the document
        must not already exist. See 'put_doc' if you want to override an
        existing document.
        If the database specifies a maximum document size and the document
        exceeds it, create will fail and raise a DocumentTooBig exception.

        :param json: The JSON document string
        :type json: str
        :param doc_id: An optional identifier specifying the document id.
        :type doc_id:
        :return:
            A deferred that will fire with the new document (A Document
            instance)
        :rtype: Deferred
        """

    def create_index(self, index_name, *index_expressions):
        """
        Create an named index, which can then be queried for future lookups.
        Creating an index which already exists is not an error, and is cheap.
        Creating an index which does not match the index_expressions of the
        existing index is an error.
        Creating an index will block until the expressions have been evaluated
        and the index generated.

        :param index_name: A unique name which can be used as a key prefix
        :type index_name: str
        :param index_expressions:
            index expressions defining the index information.
        :type index_expressions: dict

            Examples:

            "fieldname", or "fieldname.subfieldname" to index alphabetically
            sorted on the contents of a field.

            "number(fieldname, width)", "lower(fieldname)"
        """

    def delete_index(self, index_name):
        """
        Remove a named index.

        :param index_name: The name of the index we are removing
        :type index_name: str
        """

    def list_indexes(self):
        """
        List the definitions of all known indexes.

        :return: A list of [('index-name', ['field', 'field2'])] definitions.
        :rtype: Deferred
        """

    def get_from_index(self, index_name, *key_values):
        """
        Return documents that match the keys supplied.

        You must supply exactly the same number of values as have been defined
        in the index. It is possible to do a prefix match by using '*' to
        indicate a wildcard match. You can only supply '*' to trailing entries,
        (eg 'val', '*', '*' is allowed, but '*', 'val', 'val' is not.)
        It is also possible to append a '*' to the last supplied value (eg
        'val*', '*', '*' or 'val', 'val*', '*', but not 'val*', 'val', '*')

        :param index_name: The index to query
        :type index_name: str
        :param key_values: values to match. eg, if you have
                           an index with 3 fields then you would have:
                           get_from_index(index_name, val1, val2, val3)
        :type key_values: tuple
        :return: List of [Document]
        :rtype: list
        """

    def get_count_from_index(self, index_name, *key_values):
        """
        Return the count of the documents that match the keys and
        values supplied.

        :param index_name: The index to query
        :type index_name: str
        :param key_values: values to match. eg, if you have
                           an index with 3 fields then you would have:
                           get_from_index(index_name, val1, val2, val3)
        :type key_values: tuple
        :return: count.
        :rtype: int
        """

    def get_range_from_index(self, index_name, start_value, end_value):
        """
        Return documents that fall within the specified range.

        Both ends of the range are inclusive. For both start_value and
        end_value, one must supply exactly the same number of values as have
        been defined in the index, or pass None. In case of a single column
        index, a string is accepted as an alternative for a tuple with a single
        value. It is possible to do a prefix match by using '*' to indicate
        a wildcard match. You can only supply '*' to trailing entries, (eg
        'val', '*', '*' is allowed, but '*', 'val', 'val' is not.) It is also
        possible to append a '*' to the last supplied value (eg 'val*', '*',
        '*' or 'val', 'val*', '*', but not 'val*', 'val', '*')

        :param index_name: The index to query
        :type index_name: str
        :param start_values: tuples of values that define the lower bound of
            the range. eg, if you have an index with 3 fields then you would
            have: (val1, val2, val3)
        :type start_values: tuple
        :param end_values: tuples of values that define the upper bound of the
            range. eg, if you have an index with 3 fields then you would have:
            (val1, val2, val3)
        :type end_values: tuple
        :return: A deferred that will fire with a list of [Document]
        :rtype: Deferred
        """

    def get_index_keys(self, index_name):
        """
        Return all keys under which documents are indexed in this index.

        :param index_name: The index to query
        :type index_name: str
        :return:
            A deferred that will fire with a list of tuples of indexed keys.
        :rtype: Deferred
        """

    def get_doc_conflicts(self, doc_id):
        """
        Get the list of conflicts for the given document.

        :param doc_id: the document id
        :type doc_id: str

        :return:
            A deferred that will fire with a list of the document entries that
            are conflicted.
        :rtype: Deferred
        """

    def resolve_doc(self, doc, conflicted_doc_revs):
        """
        Mark a document as no longer conflicted.

        :param doc: a document with the new content to be inserted.
        :type doc: Document
        :param conflicted_doc_revs:
            A deferred that will fire with a list of revisions that the new
            content supersedes.
        :type conflicted_doc_revs: list
        """


class ISyncableStorage(Interface):
    """
    I implement methods to synchronize with a remote replica.
    """
    replica_uid = Attribute("The uid of the local replica")
    syncing = Attribute(
        "Property, True if the syncer is syncing.")
    token = Attribute("The authentication Token.")

    def sync(self):
        """
        Synchronize the local encrypted replica with a remote replica.

        This method blocks until a syncing lock is acquired, so there are no
        attempts of concurrent syncs from the same client replica.

        :param url: the url of the target replica to sync with
        :type url: str

        :return:
            A deferred that will fire with the local generation before the
            synchronisation was performed.
        :rtype: str
        """

    def stop_sync(self):
        """
        Stop the current syncing process.
        """


class ISecretsStorage(Interface):
    """
    I implement methods needed for initializing and accessing secrets, that are
    synced against the Shared Recovery Database.
    """
    secrets_file_name = Attribute(
        "The name of the file where the storage secrets will be stored")

    # XXX this used internally from secrets, so it might be good to preserve
    # as a public boundary with other components.

    # We should also probably document its interface.
    secrets = Attribute("A SoledadSecrets object containing access to secrets")

    def change_passphrase(self, new_passphrase):
        """
        Change the passphrase that encrypts the storage secret.

        :param new_passphrase: The new passphrase.
        :type new_passphrase: unicode

        :raise NoStorageSecret: Raised if there's no storage secret available.
        """



# -*- coding: utf-8 -*-
# signal.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
Signaling functions.
"""

from leap.common.events import emit_async
from leap.common.events import catalog


SOLEDAD_CREATING_KEYS = catalog.SOLEDAD_CREATING_KEYS
SOLEDAD_DONE_CREATING_KEYS = catalog.SOLEDAD_DONE_CREATING_KEYS
SOLEDAD_DOWNLOADING_KEYS = catalog.SOLEDAD_DOWNLOADING_KEYS
SOLEDAD_DONE_DOWNLOADING_KEYS = \
    catalog.SOLEDAD_DONE_DOWNLOADING_KEYS
SOLEDAD_UPLOADING_KEYS = catalog.SOLEDAD_UPLOADING_KEYS
SOLEDAD_DONE_UPLOADING_KEYS = \
    catalog.SOLEDAD_DONE_UPLOADING_KEYS
SOLEDAD_NEW_DATA_TO_SYNC = catalog.SOLEDAD_NEW_DATA_TO_SYNC
SOLEDAD_DONE_DATA_SYNC = catalog.SOLEDAD_DONE_DATA_SYNC
SOLEDAD_SYNC_SEND_STATUS = catalog.SOLEDAD_SYNC_SEND_STATUS
SOLEDAD_SYNC_RECEIVE_STATUS = catalog.SOLEDAD_SYNC_RECEIVE_STATUS


__all__ = [
    "catalog",
    "emit_async",
    "SOLEDAD_CREATING_KEYS",
    "SOLEDAD_DONE_CREATING_KEYS",
    "SOLEDAD_DOWNLOADING_KEYS",
    "SOLEDAD_DONE_DOWNLOADING_KEYS",
    "SOLEDAD_UPLOADING_KEYS",
    "SOLEDAD_DONE_UPLOADING_KEYS",
    "SOLEDAD_NEW_DATA_TO_SYNC",
    "SOLEDAD_DONE_DATA_SYNC",
    "SOLEDAD_SYNC_SEND_STATUS",
    "SOLEDAD_SYNC_RECEIVE_STATUS",
]



# -*- coding: utf-8 -*-
# send_protocol.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from zope.interface import implementer
from twisted.internet import defer
from twisted.internet import reactor
from twisted.web.iweb import IBodyProducer
from twisted.web.iweb import UNKNOWN_LENGTH


@implementer(IBodyProducer)
class DocStreamProducer(object):
    """
    A producer that writes the body of a request to a consumer.
    """

    def __init__(self, producer):
        """
        Initialize the string produer.

        :param producer: A RequestBody instance and a list of producer calls
        :type producer: (.support.RequestBody, [(function, *args)])
        """
        self.body, self.producer = producer
        self.length = UNKNOWN_LENGTH
        self.pause = False
        self.stop = False

    @defer.inlineCallbacks
    def startProducing(self, consumer):
        """
        Write the body to the consumer.

        :param consumer: Any IConsumer provider.
        :type consumer: twisted.internet.interfaces.IConsumer

        :return: A Deferred that fires when production ends.
        :rtype: twisted.internet.defer.Deferred
        """
        while self.producer and not self.stop:
            if self.pause:
                yield self.sleep(0.001)
                continue
            call = self.producer.pop(0)
            fun, args = call[0], call[1:]
            yield fun(*args)
            consumer.write(self.body.pop(1, leave_open=True))
        consumer.write(self.body.pop(0))  # close stream

    def sleep(self, secs):
        d = defer.Deferred()
        reactor.callLater(secs, d.callback, None)
        return d

    def pauseProducing(self):
        self.pause = True

    def stopProducing(self):
        self.stop = True

    def resumeProducing(self):
        self.pause = False



# -*- coding: utf-8 -*-
# fetch.py
# Copyright (C) 2015 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import json
from twisted.internet import defer
from twisted.internet import threads

from leap.soledad.client.events import SOLEDAD_SYNC_RECEIVE_STATUS
from leap.soledad.client.events import emit_async
from leap.soledad.client.http_target.support import RequestBody
from leap.soledad.common.log import getLogger
from leap.soledad.client._crypto import is_symmetrically_encrypted
from leap.soledad.common.l2db import errors
from leap.soledad.client import crypto as old_crypto

from .._document import Document
from . import fetch_protocol

logger = getLogger(__name__)


class HTTPDocFetcher(object):
    """
    Handles Document fetching from Soledad server, using HTTP as transport.
    Steps:
    * Prepares metadata by asking server for one document
    * Fetch the total on response and prepare to ask all remaining
    * (async) Documents will come encrypted.
              So we parse, decrypt and insert locally as they arrive.
    """

    # The uuid of the local replica.
    # Any class inheriting from this one should provide a meaningful attribute
    # if the sync status event is meant to be used somewhere else.

    uuid = 'undefined'
    userid = 'undefined'

    @defer.inlineCallbacks
    def _receive_docs(self, last_known_generation, last_known_trans_id,
                      ensure_callback, sync_id):
        new_generation = last_known_generation
        new_transaction_id = last_known_trans_id
        # Acts as a queue, ensuring line order on async processing
        # as `self._insert_doc_cb` cant be run concurrently or out of order.
        # DeferredSemaphore solves the concurrency and its implementation uses
        # a queue, solving the ordering.
        # FIXME: Find a proper solution to avoid surprises on Twisted changes
        self.semaphore = defer.DeferredSemaphore(1)

        metadata = yield self._fetch_all(
            last_known_generation, last_known_trans_id,
            sync_id)
        number_of_changes, ngen, ntrans = self._parse_metadata(metadata)

        # wait for pending inserts
        yield self.semaphore.acquire()

        if ngen:
            new_generation = ngen
            new_transaction_id = ntrans

        defer.returnValue([new_generation, new_transaction_id])

    def _fetch_all(self, last_known_generation,
                   last_known_trans_id, sync_id):
        # add remote replica metadata to the request
        body = RequestBody(
            last_known_generation=last_known_generation,
            last_known_trans_id=last_known_trans_id,
            sync_id=sync_id,
            ensure=self._ensure_callback is not None)
        self._received_docs = 0
        # build a stream reader with _doc_parser as a callback
        body_reader = fetch_protocol.build_body_reader(self._doc_parser)
        # start download stream
        return self._http_request(
            self._url,
            method='POST',
            body=str(body),
            content_type='application/x-soledad-sync-get',
            body_reader=body_reader)

    @defer.inlineCallbacks
    def _doc_parser(self, doc_info, content, total):
        """
        Insert a received document into the local replica, decrypting
        if necessary. The case where it's not decrypted is when a doc gets
        inserted from Server side with a GPG encrypted content.

        :param doc_info: Dictionary representing Document information.
        :type doc_info: dict
        :param content: The Document's content.
        :type idx: str
        :param total: The total number of operations.
        :type total: int
        """
        yield self.semaphore.run(self.__atomic_doc_parse, doc_info, content,
                                 total)

    @defer.inlineCallbacks
    def __atomic_doc_parse(self, doc_info, content, total):
        doc = Document(doc_info['id'], doc_info['rev'], content)
        if is_symmetrically_encrypted(content):
            content = (yield self._crypto.decrypt_doc(doc)).getvalue()
        elif old_crypto.is_symmetrically_encrypted(doc):
            content = self._deprecated_crypto.decrypt_doc(doc)
        doc.set_json(content)

        # TODO insert blobs here on the blob backend
        # FIXME: This is wrong. Using the very same SQLite connection object
        # from multiple threads is dangerous. We should bring the dbpool here
        # or find an alternative.  Deferring to a thread only helps releasing
        # the reactor for other tasks as this is an IO intensive call.
        yield threads.deferToThread(self._insert_doc_cb,
                                    doc, doc_info['gen'], doc_info['trans_id'])
        self._received_docs += 1
        user_data = {'uuid': self.uuid, 'userid': self.userid}
        _emit_receive_status(user_data, self._received_docs, total=total)

    def _parse_metadata(self, metadata):
        """
        Parse the response from the server containing the sync metadata.

        :param response: Metadata as string
        :type response: str

        :return: (number_of_changes, new_gen, new_trans_id)
        :rtype: tuple
        """
        try:
            metadata = json.loads(metadata)
            # make sure we have replica_uid from fresh new dbs
            if self._ensure_callback and 'replica_uid' in metadata:
                self._ensure_callback(metadata['replica_uid'])
            return (metadata['number_of_changes'], metadata['new_generation'],
                    metadata['new_transaction_id'])
        except (ValueError, KeyError):
            raise errors.BrokenSyncStream('Metadata parsing failed')


def _emit_receive_status(user_data, received_docs, total):
    content = {'received': received_docs, 'total': total}
    emit_async(SOLEDAD_SYNC_RECEIVE_STATUS, user_data, content)

    if received_docs % 20 == 0:
        msg = "%d/%d" % (received_docs, total)
        logger.debug("Sync receive status: %s" % msg)



# -*- coding: utf-8 -*-
# support.py
# Copyright (C) 2015 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import warnings
import json

from twisted.internet import defer
from twisted.web.client import _ReadBodyProtocol
from twisted.web.client import PartialDownloadError
from twisted.web._newclient import ResponseDone
from twisted.web._newclient import PotentialDataLoss

from leap.soledad.common.l2db import errors
from leap.soledad.common.l2db.remote import http_errors

# we want to make sure that HTTP errors will raise appropriate u1db errors,
# that is, fire errbacks with the appropriate failures, in the context of
# twisted. Because of that, we redefine the http body reader used by the HTTP
# client below.


class ReadBodyProtocol(_ReadBodyProtocol):
    """
    From original Twisted implementation, focused on adding our error
    handling and ensuring that the proper u1db error is raised.
    """

    def __init__(self, response, deferred):
        """
        Initialize the protocol, additionally storing the response headers.
        """
        _ReadBodyProtocol.__init__(
            self, response.code, response.phrase, deferred)
        self.headers = response.headers

    # ---8<--- snippet from u1db.remote.http_client, modified to use errbacks
    def _error(self, respdic):
        descr = respdic.get("error")
        exc_cls = errors.wire_description_to_exc.get(descr)
        if exc_cls is not None:
            message = respdic.get("message")
            self.deferred.errback(exc_cls(message))
        else:
            self.deferred.errback(
                errors.HTTPError(self.status, respdic, self.headers))
    # ---8<--- end of snippet from u1db.remote.http_client

    def connectionLost(self, reason):
        """
        Deliver the accumulated response bytes to the waiting L{Deferred}, if
        the response body has been completely received without error.
        """
        if reason.check(ResponseDone):

            body = b''.join(self.dataBuffer)

            # ---8<--- snippet from u1db.remote.http_client
            if self.status in (200, 201):
                self.deferred.callback(body)
            elif self.status in http_errors.ERROR_STATUSES:
                try:
                    respdic = json.loads(body)
                except ValueError:
                    self.deferred.errback(
                        errors.HTTPError(self.status, body, self.headers))
                else:
                    self._error(respdic)
            # special cases
            elif self.status == 503:
                self.deferred.errback(errors.Unavailable(body, self.headers))
            else:
                self.deferred.errback(
                    errors.HTTPError(self.status, body, self.headers))
            # ---8<--- end of snippet from u1db.remote.http_client

        elif reason.check(PotentialDataLoss):
            self.deferred.errback(
                PartialDownloadError(self.status, self.message,
                                     b''.join(self.dataBuffer)))
        else:
            self.deferred.errback(reason)


def readBody(response, protocolClass=ReadBodyProtocol):
    """
    Get the body of an L{IResponse} and return it as a byte string.

    This is a helper function for clients that don't want to incrementally
    receive the body of an HTTP response.

    @param response: The HTTP response for which the body will be read.
    @type response: L{IResponse} provider

    @return: A L{Deferred} which will fire with the body of the response.
        Cancelling it will close the connection to the server immediately.
    """
    def cancel(deferred):
        """
        Cancel a L{readBody} call, close the connection to the HTTP server
        immediately, if it is still open.

        @param deferred: The cancelled L{defer.Deferred}.
        """
        abort = getAbort()
        if abort is not None:
            abort()

    d = defer.Deferred(cancel)
    protocol = protocolClass(response, d)

    def getAbort():
        return getattr(protocol.transport, 'abortConnection', None)

    response.deliverBody(protocol)

    if protocol.transport is not None and getAbort() is None:
        warnings.warn(
            'Using readBody with a transport that does not have an '
            'abortConnection method',
            category=DeprecationWarning,
            stacklevel=2)

    return d


class RequestBody(object):
    """
    This class is a helper to generate send and fetch requests.
    The expected format is something like:
    [
    {headers},
    {entry1},
    {...},
    {entryN},
    ]
    """

    def __init__(self, **header_dict):
        """
        Creates a new RequestBody holding header information.

        :param header_dict: A dictionary with the headers.
        :type header_dict: dict
        """
        self.headers = header_dict
        self.entries = []
        self.consumed = 0

    def insert_info(self, **entry_dict):
        """
        Dumps an entry into JSON format and add it to entries list.
        Adds 'content' key on a new line if it's present.

        :param entry_dict: Entry as a dictionary
        :type entry_dict: dict
        """
        content = ''
        if 'content' in entry_dict:
            content = ',\r\n' + (entry_dict['content'] or '')
        entry = json.dumps(entry_dict) + content
        self.entries.append(entry)

    def pop(self, amount=10, leave_open=False):
        """
        Removes entries and returns it formatted and ready
        to be sent.

        :param amount: number of entries to pop and format
        :type amount: int

        :param leave_open: flag to skip stream closing
        :type amount: bool

        :return: formatted body ready to be sent
        :rtype: str
        """
        start = self.consumed == 0
        amount = min([len(self.entries), amount])
        entries = [self.entries.pop(0) for i in xrange(amount)]
        self.consumed += amount
        end = len(self.entries) == 0 if not leave_open else False
        return self.entries_to_str(entries, start, end)

    def __str__(self):
        return self.pop(len(self.entries))

    def __len__(self):
        return len(self.entries)

    def entries_to_str(self, entries=None, start=True, end=True):
        """
        Format a list of entries into the body format expected
        by the server.

        :param entries: entries to format
        :type entries: list

        :return: formatted body ready to be sent
        :rtype: str
        """
        data = ''
        if start:
            data = '[\r\n' + json.dumps(self.headers)
        data += ''.join(',\r\n' + entry for entry in entries)
        if end:
            data += '\r\n]'
        return data



# -*- coding: utf-8 -*-
# _document.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Public interfaces for adding extra client features to the generic
SoledadDocument.
"""

import weakref
import uuid

from twisted.internet import defer

from zope.interface import Interface
from zope.interface import implementer

from leap.soledad.common.document import SoledadDocument


class IDocumentWithAttachment(Interface):
    """
    A document that can have an attachment.
    """

    def set_store(self, store):
        """
        Set the store used by this file to manage attachments.

        :param store: The store used to manage attachments.
        :type store: Soledad
        """

    def put_attachment(self, fd):
        """
        Attach data to this document.

        Add the attachment to local storage, enqueue for upload.

        The document content will be updated with a pointer to the attachment,
        but the document has to be manually put in the database to reflect
        modifications.

        :param fd: A file-like object whose content will be attached to this
                   document.
        :type fd: file-like

        :return: A deferred which fires when the attachment has been added to
                 local storage.
        :rtype: Deferred
        """

    def get_attachment(self):
        """
        Return the data attached to this document.

        If document content contains a pointer to the attachment, try to get
        the attachment from local storage and, if not found, from remote
        storage.

        :return: A deferred which fires with a file like-object whose content
                 is the attachment of this document, or None if nothing is
                 attached.
        :rtype: Deferred
        """

    def delete_attachment(self):
        """
        Delete the attachment of this document.

        The pointer to the attachment will be removed from the document
        content, but the document has to be manually put in the database to
        reflect modifications.

        :return: A deferred which fires when the attachment has been deleted
                 from local storage.
        :rtype: Deferred
        """

    def get_attachment_state(self):
        """
        Return the state of the attachment of this document.

        The state is a member of AttachmentStates and is of one of NONE,
        LOCAL, REMOTE or SYNCED.

        :return: A deferred which fires with The state of the attachment of
                 this document.
        :rtype: Deferred
        """

    def is_dirty(self):
        """
        Return whether this document's content differs from the contents stored
        in local database.

        :return: A deferred which fires with True or False, depending on
                 whether this document is dirty or not.
        :rtype: Deferred
        """

    def upload_attachment(self):
        """
        Upload this document's attachment.

        :return: A deferred which fires with the state of the attachment after
                 it's been uploaded, or NONE if there's no attachment for this
                 document.
        :rtype: Deferred
        """

    def download_attachment(self):
        """
        Download this document's attachment.

        :return: A deferred which fires with the state of the attachment after
                 it's been downloaded, or NONE if there's no attachment for
                 this document.
        :rtype: Deferred
        """


class BlobDoc(object):

    # TODO probably not needed, but convenient for testing for now.

    def __init__(self, content, blob_id):

        self.blob_id = blob_id
        self.is_blob = True
        self.blob_fd = content
        if blob_id is None:
            blob_id = uuid.uuid4().get_hex()
        self.blob_id = blob_id


class AttachmentStates(object):
    NONE = 0
    LOCAL = 1
    REMOTE = 2
    SYNCED = 4


@implementer(IDocumentWithAttachment)
class Document(SoledadDocument):

    def __init__(self, doc_id=None, rev=None, json='{}', has_conflicts=False,
                 syncable=True, store=None):
        SoledadDocument.__init__(self, doc_id=doc_id, rev=rev, json=json,
                                 has_conflicts=has_conflicts,
                                 syncable=syncable)
        self.set_store(store)

    #
    # properties
    #

    @property
    def _manager(self):
        if not self.store or not hasattr(self.store, 'blobmanager'):
            raise Exception('No blob manager found to manage attachments.')
        return self.store.blobmanager

    @property
    def _blob_id(self):
        if self.content and 'blob_id' in self.content:
            return self.content['blob_id']
        return None

    def get_store(self):
        return self._store() if self._store else None

    def set_store(self, store):
        self._store = weakref.ref(store) if store else None

    store = property(get_store, set_store)

    #
    # attachment api
    #

    def put_attachment(self, fd):
        # add pointer to content
        blob_id = self._blob_id or str(uuid.uuid4())
        if not self.content:
            self.content = {}
        self.content['blob_id'] = blob_id
        # put using manager
        blob = BlobDoc(fd, blob_id)
        fd.seek(0, 2)
        size = fd.tell()
        fd.seek(0)
        return self._manager.put(blob, size)

    def get_attachment(self):
        if not self._blob_id:
            return defer.succeed(None)
        return self._manager.get(self._blob_id)

    def delete_attachment(self):
        raise NotImplementedError

    @defer.inlineCallbacks
    def get_attachment_state(self):
        state = AttachmentStates.NONE

        if not self._blob_id:
            defer.returnValue(state)

        local_list = yield self._manager.local_list()
        if self._blob_id in local_list:
            state |= AttachmentStates.LOCAL

        remote_list = yield self._manager.remote_list()
        if self._blob_id in remote_list:
            state |= AttachmentStates.REMOTE

        defer.returnValue(state)

    @defer.inlineCallbacks
    def is_dirty(self):
        stored = yield self.store.get_doc(self.doc_id)
        if stored.content != self.content:
            defer.returnValue(True)
        defer.returnValue(False)

    @defer.inlineCallbacks
    def upload_attachment(self):
        if not self._blob_id:
            defer.returnValue(AttachmentStates.NONE)

        fd = yield self._manager.get_blob(self._blob_id)
        # TODO: turn following method into a public one
        yield self._manager._encrypt_and_upload(self._blob_id, fd)
        defer.returnValue(self.get_attachment_state())

    @defer.inlineCallbacks
    def download_attachment(self):
        if not self._blob_id:
            defer.returnValue(None)
        yield self.get_attachment()
        defer.returnValue(self.get_attachment_state())
